"""
main.py -- contains classes that model scrabble game
"""
import itertools
import multiprocessing
import operator
import random

from . import config
from . import helpers

def get_move_set_generator(new_game, reference_game, move_list):
    legal_move_set = get_legal_move_set(new_game, reference_game)
    player_to_move_id = new_game.move_number % len(new_game.player_rack_list)
    player_score_list = reference_game.player_score_list_list[
        player_to_move_id
    ]

    player_move_number = new_game.move_number // len(new_game.player_rack_list)
    target_score = player_score_list[player_move_number]
    next_move_set = set(
        frozenset((tile.letter, location) for tile, location in move_set)
        for score, move_set in legal_move_set
        if score == target_score
    )

    for next_move in next_move_set:
        new_game_copy = copy_game(new_game)
        move_list_copy = move_list[:]

        player_to_move_id = (
            new_game_copy.move_number % len(new_game_copy.player_rack_list)
        )

        next_move_str = ''.join(letter for letter, location in next_move)
        new_game_copy.cheat_create_rack_word(next_move_str, player_to_move_id)
        new_game_copy.next_player_move(next_move, False)
        move_list_copy.append(next_move)

        if new_game_copy.move_number == reference_game.move_number:
            if helpers.boards_are_equivalent(reference_game.board,
                                             new_game_copy.board):
                yield move_list_copy

        else:
            yield from get_move_set_generator(new_game_copy,
                                              reference_game,
                                              move_list_copy)

def get_location_best_move_helper(argument_list):
    return get_location_best_move(*argument_list)

def get_location_best_move(game, location, word_list):
    player_to_move_id = game.move_number % len(game.player_rack_list)

    high_score = 0
    best_move = None
    for word in word_list:
        for is_vertical in [True, False]:
            temp_game = copy_game(game)
            if temp_game.place_word(word, location, is_vertical, False):
                letter_location_set = (
                    helpers.get_word_letter_location_set(
                        word,
                        location,
                        is_vertical
                    )
                )

                location_set = set(location
                                   for _, location in letter_location_set)

                if helpers.all_created_words_are_english(temp_game.board,
                                                         location_set):
                    player_score_list = (
                        temp_game.player_score_list_list[player_to_move_id]
                    )

                    word_score = player_score_list[-1]
                    if word_score > high_score:
                        best_move = (location, word, is_vertical)
                        high_score = word_score

    return high_score, best_move

def get_legal_move_set(new_game, reference_game):
    all_possible_moves_set = helpers.get_all_possible_moves_set(new_game,
                                                                reference_game)

    legal_move_set = set()
    for move_set in all_possible_moves_set:
        if helpers.move_is_legal(new_game.board,
                                 new_game.move_number,
                                 move_set):
            temp_board = copy_board(new_game.board)
            for tile, location in move_set:
                temp_board[location] = tile

            legal_move_set.add(
                (helpers.score_move(move_set, temp_board), move_set)
            )

    return legal_move_set

def copy_board(input_board):
    input_square_dict = input_board.board_square_dict

    new_board = ScrabbleBoard()
    new_square_dict = new_board.board_square_dict

    for location, square in input_square_dict.items():
        if square.tile:
            new_board_square = new_square_dict[location]
            new_board_square.letter_multiplier = square.letter_multiplier
            new_board_square.word_multiplier = square.word_multiplier
            new_board_square.tile = ScrabbleTile(
                square.tile.letter
            )

    return new_board

def copy_game(input_game):
    new_game = ScrabbleGame(len(input_game.player_rack_list))
    new_game.board = copy_board(input_game.board)
    new_game.move_number = input_game.move_number
    new_game.player_score_list_list = [
        input_player_score_list[:]
        for input_player_score_list in input_game.player_score_list_list
    ]

    new_player_rack_list = []
    for player_rack in input_game.player_rack_list:
        new_rack = []
        for tile in player_rack:
            new_rack.append(ScrabbleTile(tile.letter))

        new_player_rack_list.append(new_rack)

    new_game.player_rack_list = new_player_rack_list

    return new_game

def get_move_set_notation(move_set, reference_game):
    new_game = ScrabbleGame(len(reference_game.player_rack_list))
    word_notation_list_list = [
        [] for _ in range(len(reference_game.player_rack_list))
    ]

    for move in move_set:
        player_to_move_id = (
            new_game.move_number % len(new_game.player_rack_list)
        )

        move_location_set = set(location for _, location in move)
        rack_word = ''.join([letter for letter, _ in move])
        new_game.cheat_create_rack_word(rack_word, player_to_move_id)

        player_words_notation_list = (
            word_notation_list_list[player_to_move_id]
        )

        notation_word_list = []
        new_game.next_player_move(move, False)
        word_set = helpers.get_word_set(new_game.board, move_location_set)
        for word_location_set in word_set:
            if word_location_set:
                notation_word_list.append(
                    helpers.get_move_word(word_location_set,
                                          move_location_set,
                                          new_game)
                )

        player_words_notation_list.append(notation_word_list)

    return word_notation_list_list

def get_new_tile_bag():
    return [ScrabbleTile(letter=letter)
            for letter, magnitude in config.LETTER_DISTRIBUTION_DICT.items()
            for _ in range(magnitude)]

def read_input_file(input_filename):
    board_character_array, player_score_list_list = helpers.load_file(
        input_filename
    )

    num_players = len(player_score_list_list)
    game = ScrabbleGame(num_players)
    game.player_score_list_list = player_score_list_list
    game.player_rack_list = [[] for _ in range(num_players)]
    game.tile_bag = []
    game.move_number = sum(1 for player_score_list in player_score_list_list
                             for _ in player_score_list)

    for row_number, row in enumerate(board_character_array):
        for column_number, letter in enumerate(row):
            if letter:
                column_letter = chr(ord('a') + column_number)
                this_location = (column_letter, row_number + 1)
                game.board[this_location] = ScrabbleTile(letter)

    return game

def initialize_new_board_square_dict():
    initial_board_square_dict = {}
    for column in config.BOARD_CODE_DICT:
        for row in range(1, config.BOARD_NUM_ROWS + 1):
            location = (column, row)

            word_multiplier = config.WORD_SCORE_MULT_LOCATION_DICT.get(
                location,
                1
            )

            letter_multiplier = config.LETTER_SCORE_MULT_LOCATION_DICT.get(
                location,
                1
            )

            initial_board_square_dict[location] = BoardSquare(
                tile=None,
                word_multiplier=word_multiplier,
                letter_multiplier=letter_multiplier
            )

    return initial_board_square_dict

def recover_game(input_filename):
    reference_game = read_input_file(input_filename)
    new_game = ScrabbleGame(
        len(reference_game.player_rack_list)
    )

    move_set_generator = get_move_set_generator(new_game,
                                                reference_game,
                                                [])

    move_set_list = list(move_set_generator)

    notated_move_set_list = [
        get_move_set_notation(move_set, reference_game)
        for move_set in move_set_list
    ]

    return notated_move_set_list


class ScrabbleTile(object):
    def __init__(self, letter):
        self.letter = letter
        self.point_value = config.LETTER_POINT_VALUES_DICT[letter]

    def __repr__(self):
        return self.letter


class BoardSquare(object):
    def __init__(self, tile, letter_multiplier, word_multiplier):
        self.tile = tile
        self.letter_multiplier = letter_multiplier
        self.word_multiplier = word_multiplier

    def __repr__(self):
        if self.tile:
            return str(self.tile)
        else:
            return config.BLANK_SQUARE_CHARACTER


class ScrabbleBoard(object):
    def __init__(self):
        self.board_square_dict = initialize_new_board_square_dict()

        center_row = (config.BOARD_NUM_ROWS // 2) + 1
        center_column = chr(
            (config.BOARD_NUM_COLUMNS // 2) + config.LOWER_COLUMN_INT_BOUND
        )

        self.start_square_location = (center_column, center_row)

    def __getitem__(self, key):
        return self.board_square_dict.get(key).tile

    def __setitem__(self, key, value):
        self.board_square_dict[key].tile = value

    def __repr__(self):
        square_letter_gen = (
            str(square)
            for location, square in sorted(self.board_square_dict.items())
        )

        board_array_first_row = (
            [' ', ' '] + sorted(list(config.BOARD_CODE_DICT.keys()))
        )

        board_array = [board_array_first_row]  # Column labels
        board_array.extend(
            [' ' for _ in range(config.BOARD_NUM_COLUMNS+1)]
            for _ in range(config.BOARD_NUM_ROWS)
        )

        for i in range(1, config.BOARD_NUM_ROWS + 1):
            board_array[i][0] = str(i)  # Row labels
            if i < 10:
                board_array[i][0] += ' '  # Pad single digit numbers with space

            for j in range(1, config.BOARD_NUM_COLUMNS + 1):
                board_array[j][i] = next(square_letter_gen)  # swap x and y

        center_row_num = (config.BOARD_NUM_ROWS // 2) + 1
        center_column_num = (config.BOARD_NUM_COLUMNS // 2) + 1
        start_char = config.START_SQUARE_CHARACTER
        blank_char = config.BLANK_SQUARE_CHARACTER

        if board_array[center_column_num][center_row_num] == blank_char:
            board_array[center_column_num][center_row_num] = start_char

        return_line_list = [''.join(row) for row in board_array]
        return_str = '\n'.join(return_line_list)

        return return_str


class ScrabbleGame(object):
    def __init__(self, num_players):
        self.tile_bag = get_new_tile_bag()
        self.player_rack_list = self._get_new_player_rack_list(num_players)
        self.board = ScrabbleBoard()
        self.player_score_list_list = [[] for _ in range(num_players)]
        self.move_number = 0

    def __repr__(self):
        player_score_str_list = []
        for i, player_score_list in enumerate(self.player_score_list_list):
            player_score_str_list.append(
                'Player {player_number}: {score}'.format(
                    player_number=(i + 1),
                    score=sum(player_score_list)
                )
            )

        player_scores_str = '\n'.join(player_score_str_list)
        player_to_move_id, _ = helpers.get_current_player_data(
            self.move_number,
            self.player_rack_list
        )

        return (
            '{board}\n'
            '{player_rack_list}\n'
            'Moves played: {move_number}\n'
            'Player {player_to_move}\'s move\n'
            '{tiles_remaining} tiles remain in bag\n'
            '{player_scores_str}'
        ).format(
            board=str(self.board),
            player_rack_list=self.player_rack_list,
            move_number=self.move_number,
            player_to_move=player_to_move_id + 1,
            tiles_remaining=len(self.tile_bag),
            player_scores_str=player_scores_str
        )

    def exchange(self, letter_list):
        if (len(self.tile_bag) < config.PLAYER_RACK_SIZE or
                len(letter_list) > config.PLAYER_RACK_SIZE):
            return False
        else:
            _, player_rack = helpers.get_current_player_data(
                self.move_number,
                self.player_rack_list
            )

            player_letter_list = [tile.letter for tile in player_rack]

            if helpers.move_is_sublist(letter_list, player_letter_list):
                self._perform_bag_exchange(letter_list, player_rack)
                self.move_number += 1
                return True
            else:
                return False

    def place_word(self, word, start_location, is_vertical_move,
                   allow_challenge=True):
        letter_location_set = helpers.get_word_letter_location_set(
            word,
            start_location,
            is_vertical_move
        )

        return self.next_player_move(letter_location_set, allow_challenge)

    def next_player_move(self, letter_location_set, allow_challenge=True):
        player_to_move_id, player_rack = helpers.get_current_player_data(
            self.move_number,
            self.player_rack_list
        )

        is_legal_move = helpers.move_is_legal(self.board,
                                              self.move_number,
                                              letter_location_set,
                                              player_rack)

        if is_legal_move:
            if allow_challenge and helpers.move_successfully_challenged():
                letter_location_set = set()

            for move_letter, board_location in letter_location_set:
                tile_index = helpers.get_rack_tile_index(player_rack,
                                                         move_letter)

                tile_obj = player_rack.pop(tile_index)
                self.board[board_location] = tile_obj

            move_score = helpers.score_move(letter_location_set, self.board)
            self.player_score_list_list[player_to_move_id].append(move_score)
            self._refill_player_rack(player_rack)
            self._cancel_bonus_squares(letter_location_set)

            if len(player_rack) == 0 and len(self.tile_bag) == 0:  # Final move
                self.conclude_game(empty_rack_id=player_to_move_id)

            self.move_number += 1
            return True
        else:
            return False

    def cheat_create_rack_word(self, word, player_id):
        player_rack = self.player_rack_list[player_id]
        for character in word:
            tile = ScrabbleTile(letter=character)
            player_rack.append(tile)

    def conclude_game(self, empty_rack_id=None):
        if empty_rack_id:
            last_move_score_list = helpers.score_playing_out(
                self.player_rack_list,
                empty_rack_id
            )

            for i, last_move_score in enumerate(last_move_score_list):
                self.player_score_list_list[i].append(last_move_score)

        player_score_total_list = [
            sum(player_score_list)
            for player_score_list in self.player_score_list_list
        ]

        winning_player_id, winning_player_score = max(
            enumerate(player_score_total_list),
            key=operator.itemgetter(1)
        )

        print(
            'Game Over! Player {} wins with a score of {}'.format(
                winning_player_id + 1,
                winning_player_score
            )
        )

    def get_best_move(self):
        player_to_move_id = self.move_number % len(self.player_rack_list)
        player_rack = self.player_rack_list[player_to_move_id]
        player_letter_list = [tile.letter for tile in player_rack]

        word_list = []
        for i in range(1, config.PLAYER_RACK_SIZE + 1):
            for this_list in itertools.permutations(player_letter_list, i):
                this_word = ''.join(this_list)
                word_list.append(this_word)

        input_arguments_list = [
            (self, location, word_list)
            for location in sorted(self.board.board_square_dict)
        ]

        process_pool = multiprocessing.Pool(config.NUM_PROCESSING_CORES)
        result_list = process_pool.map(get_location_best_move_helper,
                                       input_arguments_list)

        return max(result_list)

    def _get_new_player_rack_list(self, num_players):
        player_rack_list = []

        for _ in range(num_players):
            this_rack = []
            for _ in range(config.PLAYER_RACK_SIZE):
                this_tile = self._draw_random_tile()
                this_rack.append(this_tile)

            player_rack_list.append(this_rack)

        return player_rack_list

    def _draw_random_tile(self):
        random_index = random.randrange(0, len(self.tile_bag))
        selected_tile = self.tile_bag.pop(random_index)

        return selected_tile

    def _refill_player_rack(self, player_rack):
        while len(player_rack) < config.PLAYER_RACK_SIZE:
            if self.tile_bag:
                tile = self._draw_random_tile()
                player_rack.append(tile)
            else:
                break

    def _cancel_bonus_squares(self, letter_location_set):
        for _, location in letter_location_set:
            square = self.board.board_square_dict[location]
            square.letter_multiplier = 1
            square.word_multiplier = 1

    def _perform_bag_exchange(self, letter_list, player_rack):
        exchange_tile_list = []
        for letter in letter_list:
            for tile in player_rack:
                if tile.letter == letter:
                    exchange_tile_list.append(tile)
                    player_rack.remove(tile)

        for _ in range(len(letter_list)):
            new_tile = self._draw_random_tile()
            player_rack.append(new_tile)

        self.tile_bag.extend(exchange_tile_list)



"""
helpers.py -- contains all side-effect free functions which are used by classes
              in main.py
"""
import collections
import itertools
import json

from . import config

def get_all_possible_moves_set(new_game, reference_game):
    game_tile_location_set = get_all_board_tiles(new_game)
    reference_tile_location_set = get_all_board_tiles(reference_game)

    search_set = set()
    for reference_tile, reference_location in reference_tile_location_set:
        flag = True
        for game_tile, game_location in game_tile_location_set:
            if (game_tile.letter == reference_tile.letter and
                    game_location == reference_location):
                flag = False

        if flag:
            search_set.add((reference_tile, reference_location))

    return get_combinations(search_set)

def get_all_board_tiles(game):
    return set((square_tuple[1].tile, square_tuple[0])  # tile then location
               for square_tuple in game.board.board_square_dict.items()
               if square_tuple[1].tile)

def get_combinations(input_iterable):
    combination_set = set()

    for column_letter in config.BOARD_CODE_DICT:
        column_tile_set = frozenset((tile, location)
                                    for tile, location in input_iterable
                                    if location[0] == column_letter)

        for i in range(1, config.PLAYER_RACK_SIZE + 1):
            for this_set in itertools.combinations(column_tile_set, i):
                combination_set.add(this_set)

    for row_number in range(1, config.BOARD_NUM_ROWS + 1):
        row_tile_set = frozenset((tile, location)
                                 for tile, location in input_iterable
                                 if location[1] == row_number)

        for i in range(1, config.PLAYER_RACK_SIZE + 1):
            for this_set in itertools.combinations(row_tile_set, i):
                combination_set.add(this_set)

    return combination_set

def load_file(input_filename):
    with open(input_filename, 'r') as filehandle:
        input_str = filehandle.read()

    input_dict = json.loads(input_str)
    board_character_array = input_dict['board']
    player_score_list_list = input_dict['scores']

    return board_character_array, player_score_list_list

def boards_are_equivalent(board_1, board_2):
    return str(board_1) == str(board_2)

def all_created_words_are_english(board, letter_location_set):
    word_set = get_word_set(board, letter_location_set)

    for word_location_set in word_set:
        if word_location_set:
            this_word_str = ''
            for location in sorted(word_location_set):
                square = board.board_square_dict[location]
                this_word_str += str(square.tile.letter)

            if this_word_str.lower() not in config.ENGLISH_DICTIONARY_SET:
                return False

    return True

def get_move_word(word_location_set, move_location_set, game):
    move_word = ''
    word_location_list = sorted(word_location_set)
    notation_location = word_location_list[0]
    parens_flag = False

    for location in word_location_list:
        tile = game.board[location]
        if tile:
            if location not in move_location_set:
                if not parens_flag:
                    move_word += '('
                    parens_flag = True
            else:
                if parens_flag:
                    move_word += ')'
                    parens_flag = False

            move_word += tile.letter

    if parens_flag:
        move_word += ')'

    return notation_location, move_word

def score_playing_out(player_rack_list, empty_rack_id):
    final_move_score_list = [0 for _ in range(len(player_rack_list))]
    all_rack_points = 0

    for i, player_rack in enumerate(player_rack_list):
        rack_point_total = sum(tile.point_value for tile in player_rack)
        final_move_score_list[i] += (-1 * rack_point_total)
        all_rack_points += rack_point_total

    final_move_score_list[empty_rack_id] += all_rack_points

    return final_move_score_list

def score_move(letter_location_set, board):
    move_location_set = set(location for _, location in letter_location_set)
    word_set = get_word_set(board, move_location_set)
    total_score = get_word_set_total_score(board,
                                           word_set,
                                           len(move_location_set))

    return total_score

def decrement_letter(character):
    return chr(config.LETTER_CODE_DICT[character] - 1)

def increment_letter(character):
    return chr(config.LETTER_CODE_DICT[character] + 1)

def get_current_player_data(move_number, player_rack_list):
    num_players = len(player_rack_list)
    player_to_move_id = move_number % num_players
    player_rack = player_rack_list[player_to_move_id]

    return player_to_move_id, player_rack

def get_word_letter_location_set(word, start_location, is_vertical_move):
    letter_location_set = set()
    next_location_func = get_next_location_function(
        use_positive_seek=True,
        use_vertical_words=is_vertical_move
    )

    current_location = start_location
    word_iterator = iter(word)
    for character in word_iterator:
        if character == '(':  # characters in parenthesis are existing tiles
            character = next(word_iterator, None)
            while character != ')':
                current_location = next_location_func(current_location)
                character = next(word_iterator, None)

            character = next(word_iterator, None)

        if character:
            letter_location_set.add((character, current_location))
            current_location = next_location_func(current_location)

    return letter_location_set

def move_is_legal(board, move_number, letter_location_set, player_rack=None):
    letter_list = [letter for letter, _ in letter_location_set]
    location_set = set(location for _, location in letter_location_set)

    return_bool = (
        move_is_rack_size_or_less(location_set) and
        move_does_not_misalign_tiles(location_set) and
        move_is_not_out_of_bounds(location_set) and
        all_move_tiles_connected(board, location_set) and
        move_does_not_stack_tiles(letter_list, location_set) and
        move_does_not_cover_tiles(board, location_set) and
        move_touches_tile(move_number, board, location_set)
    )

    if player_rack:
        player_rack_letter_list = [tile.letter for tile in player_rack]
        return_bool = return_bool and move_is_sublist(letter_list,
                                                      player_rack_letter_list)

    return return_bool

def move_touches_tile(move_number, board, location_set):
    if move_number == 0:
        if board.start_square_location in location_set:
            return True
    else:
        for this_location in location_set:
            if location_touches_tile(board, this_location):
                return True

    return False

def move_is_sublist(letter_list_1, letter_list_2):
    letter_counter_1 = collections.Counter(letter_list_1)
    letter_counter_2 = collections.Counter(letter_list_2)
    for letter, cardinality in letter_counter_1.items():
        if cardinality > letter_counter_2[letter]:
            return False

    return True

def move_does_not_cover_tiles(board, location_set):
    for location in location_set:
        if board[location]:
            return False

    return True

def all_move_tiles_connected(board, location_set):
    column_list = [column for column, _ in location_set]
    row_list = [row for _, row in location_set]
    move_is_vertical = (len(set(column_list)) == 1)

    if move_is_vertical:
        this_column = column_list[0]
        for this_row in range(min(row_list), max(row_list) + 1):
            this_tile = board[(this_column, this_row)]
            if not (this_tile or (this_column, this_row) in location_set):
                return False
    else:
        column_range = range(
            config.LETTER_CODE_DICT[min(column_list)],
            config.LETTER_CODE_DICT[max(column_list)] + 1
        )

        this_row = row_list[0]
        for this_column_num in column_range:
            this_column = chr(this_column_num)
            this_tile = board[(this_column, this_row)]
            if not (this_tile or (this_column, this_row) in location_set):
                return False

    return True

def move_does_not_misalign_tiles(location_set):
    column_set = set(column for column, _ in location_set)
    row_set = set(row for _, row in location_set)

    return len(column_set) == 1 or len(row_set) == 1

def move_does_not_stack_tiles(letter_list, location_set):
    return len(letter_list) == len(location_set)

def move_is_rack_size_or_less(location_set):
    return len(location_set) <= config.PLAYER_RACK_SIZE

def move_is_not_out_of_bounds(location_set):
    for location in location_set:
        if location_is_out_of_bounds(location):
            return False

    return True

def move_successfully_challenged():
    response = input('Challenge successful (Y/N)')
    if response.upper() == 'Y':
        return True
    elif response.upper() == 'N':
        return False
    else:
        return move_successfully_challenged()

def location_is_out_of_bounds(location):
    column, row = location
    return (
        config.LETTER_CODE_DICT[column] < config.LOWER_COLUMN_INT_BOUND or
        config.LETTER_CODE_DICT[column] > config.UPPER_COLUMN_INT_BOUND or
        row < 1 or
        row > config.BOARD_NUM_ROWS
    )

def location_touches_tile(board, location):
    adjacent_location_set = get_adjacent_location_set(location)
    for adjacent_location in adjacent_location_set:
        if board[adjacent_location]:
            return True

    return False

def get_rack_tile_index(player_rack, move_letter):
    for i, rack_tile in enumerate(player_rack):
        if rack_tile.letter == move_letter:
            return i

    return None

def get_next_location_function(use_positive_seek, use_vertical_words):
    if use_vertical_words and use_positive_seek:
        return lambda x: (x[0], x[1] + 1)
    elif use_vertical_words and not use_positive_seek:
        return lambda x: (x[0], x[1] - 1)
    elif not use_vertical_words and use_positive_seek:
        return lambda x: (increment_letter(x[0]), x[1])
    elif not use_vertical_words and not use_positive_seek:
        return lambda x: (decrement_letter(x[0]), x[1])

def get_adjacent_location_set(location):
    column, row = location

    adjacent_location_set = set([(increment_letter(column), row),
                                 (decrement_letter(column), row),
                                 (column, row + 1),
                                 (column, row - 1)])
    # Board boundary check
    bad_location_set = set(location for location in adjacent_location_set
                           if location_is_out_of_bounds(location))

    return adjacent_location_set - bad_location_set

def get_word_location_set(board, initial_location, use_vertical_words):
    word_location_set = set()

    for use_positive_seek in [True, False]:  # Search tiles in 2 directions:
        current_location = initial_location  # either up/down or left/right
        current_tile = board[current_location]
        next_location_func = get_next_location_function(use_positive_seek,
                                                        use_vertical_words)

        while current_tile:
            word_location_set.add(current_location)
            current_location = next_location_func(current_location)
            if location_is_out_of_bounds(current_location):
                current_tile = None
            else:
                current_tile = board[current_location]

    if len(word_location_set) > 1:  # Must be at least 2 letters to be a word
        return frozenset(word_location_set)  # Must be hashable so we can have
    else:                                    # a set of frozensets
        return frozenset()

def get_word_set(board, move_location_set):
    return set(
        get_word_location_set(board, location, vertical_bool)
        for vertical_bool in [True, False]  # Search for vertical words created
        for location in move_location_set   # then for horizontal words
    )

def get_word_set_total_score(board, word_set, num_move_locations):
    total_score = 0
    word_score = 0

    for word_location_set in word_set:
        word_score = 0
        word_multiplier = 1
        this_word_str = ''

        for location in word_location_set:
            square = board.board_square_dict[location]
            word_multiplier *= square.word_multiplier
            word_score += square.tile.point_value * square.letter_multiplier
            this_word_str += square.tile.letter

        word_score *= word_multiplier
        total_score += word_score

    if num_move_locations == config.PLAYER_RACK_SIZE:
        total_score += config.BINGO_SCORE

    return total_score



"""
config.py -- contains all constants needed by scrabble module
"""
import json
import os

BOARD_NUM_COLUMNS = 15
BOARD_NUM_ROWS = 15
PLAYER_RACK_SIZE = 7
BINGO_SCORE = 50
DICTIONARY_FILENAME = 'dictionary.json'
RESOURCE_PACKAGE = 'scrabble'

START_SQUARE_CHARACTER = '★'
BLANK_SQUARE_CHARACTER = '_'
LOWEST_COLUMN_CHARACTER = 'a'

NUM_PROCESSING_CORES = 8

WORD_SCORE_MULT_LOCATION_DICT = {  # Lowercase letters are used to distinguish
    ('b', 2): 2,                   # locations from tile letter symbols
    ('b', 14): 2,
    ('c', 3): 2,
    ('c', 13): 2,
    ('d', 4): 2,
    ('d', 12): 2,
    ('e', 5): 2,
    ('e', 11): 2,
    ('h', 8): 2,
    ('k', 5): 2,
    ('k', 11): 2,
    ('l', 4): 2,
    ('l', 12): 2,
    ('m', 4): 2,
    ('m', 13): 2,
    ('n', 2): 2,
    ('n', 14): 2,
    ('a', 1): 3,
    ('a', 8): 3,
    ('a', 15): 3,
    ('h', 1): 3,
    ('h', 15): 3,
    ('o', 1): 3,
    ('o', 8): 3,
    ('o', 15): 3
}

LETTER_SCORE_MULT_LOCATION_DICT = {
    ('a', 4): 2,
    ('a', 12): 2,
    ('c', 7): 2,
    ('c', 9): 2,
    ('d', 1): 2,
    ('d', 8): 2,
    ('d', 15): 2,
    ('g', 3): 2,
    ('g', 7): 2,
    ('g', 9): 2,
    ('g', 13): 2,
    ('h', 4): 2,
    ('h', 12): 2,
    ('i', 3): 2,
    ('i', 7): 2,
    ('i', 9): 2,
    ('i', 13): 2,
    ('l', 1): 2,
    ('l', 8): 2,
    ('l', 15): 2,
    ('m', 7): 2,
    ('m', 9): 2,
    ('o', 4): 2,
    ('m', 12): 2,
    ('b', 6): 3,
    ('b', 10): 3,
    ('f', 2): 3,
    ('f', 6): 3,
    ('f', 10): 3,
    ('f', 14): 3,
    ('j', 2): 3,
    ('j', 6): 3,
    ('j', 10): 3,
    ('j', 14): 3,
    ('n', 6): 3,
    ('n', 10): 3
}

LETTER_POINT_VALUES_DICT = {  # Captial letters are used to distinguish tile
    '*': 0,                   # letter symbols from locations
    'A': 1,
    'B': 3,
    'C': 3,
    'D': 2,
    'E': 1,
    'F': 4,
    'G': 2,
    'H': 4,
    'I': 1,
    'J': 8,
    'K': 5,
    'L': 1,
    'M': 3,
    'N': 1,
    'O': 1,
    'P': 3,
    'Q': 10,
    'R': 1,
    'S': 1,
    'T': 1,
    'U': 1,
    'V': 4,
    'W': 4,
    'X': 8,
    'Y': 4,
    'Z': 10
}

LETTER_DISTRIBUTION_DICT = {
    '*': 2,
    'A': 9,
    'B': 2,
    'C': 2,
    'D': 4,
    'E': 12,
    'F': 2,
    'G': 3,
    'H': 2,
    'I': 9,
    'J': 1,
    'K': 1,
    'L': 4,
    'M': 2,
    'N': 6,
    'O': 8,
    'P': 2,
    'Q': 1,
    'R': 6,
    'S': 4,
    'T': 6,
    'U': 4,
    'V': 2,
    'W': 2,
    'X': 1,
    'Y': 2,
    'Z': 1
}

# Derived constants
CONTAINING_FOLDER = os.path.dirname(__file__)
FULL_DICTIONARY_PATH = os.path.join(CONTAINING_FOLDER, DICTIONARY_FILENAME)

with open(FULL_DICTIONARY_PATH) as filehandle:
    FULL_DICTIONARY_STR = filehandle.read()

ENGLISH_DICTIONARY_SET = set(json.loads(FULL_DICTIONARY_STR))

LOWER_COLUMN_INT_BOUND = ord(LOWEST_COLUMN_CHARACTER)
UPPER_COLUMN_INT_BOUND = LOWER_COLUMN_INT_BOUND + BOARD_NUM_COLUMNS - 1

LETTER_CODE_DICT = {
    chr(code_point): code_point
    for code_point in range(LOWER_COLUMN_INT_BOUND - 40,
                            UPPER_COLUMN_INT_BOUND + 40)
}

BOARD_CODE_DICT = {
    chr(code_point): code_point
    for code_point in range(LOWER_COLUMN_INT_BOUND,
                            UPPER_COLUMN_INT_BOUND + 1)
}



from setuptools import setup

setup(name='scrabble',
      version='1.0',
      description='Scrabble game with move recovery and best move analysis',
      url='https://github.com/benjamincrom/scrabble',
      author='Benjamin B. Crom',
      author_email='benjamincrom@gmail.com',
      include_package_data=True,
      license='MIT',
      packages=['scrabble'],
      setup_requires=['pytest-runner'],
      scripts=['bin/recover_scrabble_game'],
      tests_require=['pytest'],
      zip_safe=False)
